import sys
import traceback
import json
from os.path import split, splitext

from typing import Sequence
from typing import TextIO, BinaryIO
from typing import Optional

from catas import parsers
from catas.parsers import FileType
from catas.parsers import ParseError

from catas.cli import MyArgumentError
from catas.train.cli import cli

from catas.count import cazy_counts_multi
from catas.count import HMMError

from catas.model import NomenclatureClass, HMMLengths, Model

from catas import data

from catas import __email__
from catas.exitcodes import (
    EXIT_KEYBOARD, EXIT_UNKNOWN, EXIT_CLI, EXIT_INPUT_FORMAT, EXIT_SYSERR
)


def runner(
    nomenclatures_handle: Optional[TextIO],
    classes: str,
    inhmms: TextIO,
    inhandles: Sequence[TextIO],
    outhandle: BinaryIO,
    labels: Sequence[str],
    file_format: FileType,
):
    """ Runs the pipeline. """

    if nomenclatures_handle is None:
        nomenclatures = data.nomenclatures()
    else:
        nomenclatures = json.load(nomenclatures_handle)

        required_nomenclatures = {"nomenclature1", "nomenclature2",
                                  "nomenclature3"}
        if set(nomenclatures.keys()) != required_nomenclatures:
            raise ValueError("The nomenclatures json file has invalid keys.")

    with open(classes, newline='') as handle:
        class_labels = NomenclatureClass.from_tsv(handle)

    # This checks if there are any labels in input filenames that aren't in
    # the classes tsv, and vice-versa. We want 1-1.
    if len(set(c.label for c in class_labels)
           .symmetric_difference(labels)) != 0:
        print(set(c.label for c in class_labels)
           .symmetric_difference(labels))
        raise ValueError("The file labels and the class labels are different.")

    class_labels_nomenclatures = {
        "nomenclature1": {t.nomenclature1 for t in class_labels},
        "nomenclature2": {t.nomenclature2 for t in class_labels},
        "nomenclature3": {t.nomenclature3 for t in class_labels},
    }

    for nom, nom_class_set in class_labels_nomenclatures.items():
        if len(nom_class_set.symmetric_difference(nomenclatures[nom])) != 0:
            raise ValueError(f"The nomenclatures and class files for {nom} "
                             "don't have the same classes.")

    hmm_lengths = HMMLengths.read_hmm(inhmms)

    parsed = [
        parsers.parse(h, format=file_format, hmm_lens=hmm_lengths)
        for h
        in inhandles
    ]

    required_cols = sorted(hmm_lengths.keys())

    # Columns
    counts = cazy_counts_multi(parsed, labels, required_cols)
    model = Model.fit(counts, class_labels, nomenclatures, hmm_lengths)

    # The model_mean and model_components together make up the model.
    model.write(outhandle)
    return


def main():  # noqa
    try:
        args = cli(prog=sys.argv[0], args=sys.argv[1:])
    except MyArgumentError as e:
        print(e.message, file=sys.stderr)
        sys.exit(e.errno)

    infile_names = [splitext(split(f.name)[-1])[0] for f in args.infile]

    if args.labels is None:
        labels = infile_names
    elif len(args.labels) != len(args.infile):
        msg = (
            "argument labels and inhandles: \n"
            "When specified, the number of labels must be the same as the "
            "number of input files. Exiting.\n"
        )

        print(msg, file=sys.stderr)
        sys.exit(EXIT_CLI)
    else:
        labels = args.labels

    try:
        runner(
            args.nomenclatures,
            args.classes,
            args.hmms,
            args.infile,
            args.outhandle,
            labels,
            args.file_format,
        )

    except ParseError as e:
        if e.line is not None:
            header = "Failed to parse file <{}> at line {}.\n".format(
                e.filename, e.line)
        else:
            header = "Failed to parse file <{}>.\n".format(e.filename)

        print("{}\n{}".format(header, e.message), file=sys.stderr)
        sys.exit(EXIT_INPUT_FORMAT)

    except HMMError as e:
        msg = (
            "Encountered an hmm that wasn't present in the training data.\n"
            f"Offending HMMs were: {', '.join(e.hmms)}"
        )
        print(msg, file=sys.stderr)
        sys.exit(EXIT_INPUT_FORMAT)
        pass

    except OSError as e:
        msg = (
            "Encountered a system error.\n"
            "We can't control these, and they're usually related to your OS.\n"
            "Try running again.\n"
        )
        print(msg, file=sys.stderr)
        print(e.strerror, file=sys.stderr)
        sys.exit(EXIT_SYSERR)

    except MemoryError:
        msg = (
            "Ran out of memory!\n"
            "Catastrophy shouldn't use much RAM, so check other "
            "processes and try running again."
        )
        print(msg, file=sys.stderr)
        sys.exit(EXIT_SYSERR)

    except KeyboardInterrupt:
        print("Received keyboard interrupt. Exiting.", file=sys.stderr)
        sys.exit(EXIT_KEYBOARD)

    except Exception as e:
        msg = (
            "I'm so sorry, but we've encountered an unexpected error.\n"
            "This shouldn't happen, so please file a bug report with the "
            "authors.\nWe will be extremely grateful!\n\n"
            "You can email us at {}.\n"
            "Alternatively, you can file the issue directly on the repo "
            "<https://github.com/ccdmb/catastrophy/issues>\n\n"
            "Please attach a copy of the following message:"
        ).format(__email__)
        print(e, file=sys.stderr)
        traceback.print_exc(file=sys.stderr)
        sys.exit(EXIT_UNKNOWN)
    return



import re

from typing import TextIO
from typing import List, Sequence
from typing import Optional

from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq
from Bio.Data.IUPACData import protein_letters

GAP_REGEX = re.compile(r"-\s")
DEGEN_REGEX = re.compile("BZJUO*", flags=re.IGNORECASE)
IUPAC_REGEX = re.compile(f"[^{protein_letters}X*]", flags=re.IGNORECASE)


class FastaError(Exception):

    def __init__(self, messages: Sequence[str]):
        self.messages = list(messages)
        return


class FastaBadCharError(Exception):

    def __init__(
        self,
        filename: Optional[str],
        id: str,
        invalid_chars: str,
    ):
        self.filename = filename
        self.id = id
        self.invalid_chars = invalid_chars

    def str(self) -> str:
        s = ["BAD CHARACTER:"]
        if self.filename is not None:
            s.append(self.filename)

        s.extend(["ID", self.id, "-", f"'{self.invalid_chars}'"])
        return " ".join(s)


def sanitise_fasta(handle: TextIO, correct: bool) -> List[SeqRecord]:
    out = []

    seqs = SeqIO.parse(handle, format="fasta")
    errors = []
    for seq in seqs:
        if correct:
            newseq = str(seq.seq.rstrip("*")).upper()
            newseq = GAP_REGEX.sub("", newseq)
            newseq = DEGEN_REGEX.sub("X", newseq)
            seq.seq = Seq(newseq)

        match = IUPAC_REGEX.search(str(seq.seq))
        if match is None:
            out.append(seq)
        else:
            all_matches = "".join(sorted(set(
                IUPAC_REGEX.findall(str(seq.seq))
            )))
            if hasattr(handle, "name"):
                n: Optional[str] = handle.name
            else:
                n = None

            errors.append(
                FastaBadCharError(
                    n,
                    seq.id,
                    all_matches
                )
            )

    if len(errors) > 0:
        messages = [e.str() for e in errors]
        raise FastaError(messages)
    elif len(out) == 0:
        if hasattr(handle, "name"):
            n = handle.name
        else:
            n = ""
        raise FastaError([
            f"File {n} appears to be empty or is not Fasta formatted."
        ])

    return out



import argparse
import sys

from typing import List

from catas import __version__

from catas.parsers import FileType
from catas.cli import MyArgumentParser


def cli(prog: str, args: List[str]) -> argparse.Namespace:

    parser = MyArgumentParser(description="Train models for catastrophy.")

    parser.add_argument(
        "classes",
        type=str,
        help=(
            "A tsv file containing the columns: 'label', 'nomenclature1', "
            "'nomenclature2', nomenclature3'"
        )
    )

    parser.add_argument(
        "hmms",
        type=argparse.FileType('r'),
        help=(
            "Path to the hmmer formatted database. "
            "We use this to get HMM lengths."
        )
    )

    parser.add_argument(
        "infile",
        nargs="+",
        type=argparse.FileType('r'),
        help=(
            "Path to the input file output by HMMER or formatted by dbCAN "
            "`hmmscan-parser.sh`. You can specify more than one file by "
            "separating them with a space."
        )
    )

    parser.add_argument(
        "-f", "--format",
        dest="file_format",
        type=FileType.from_string,
        default=FileType.hmmer_text,
        choices=list(FileType),
        help=(
            "The format that the input is provided in. If multiple files are "
            "specified, all input must be in the same format. HMMER raw "
            "(hmmer_text, default) and domain table (hmmer_domtab) formatted "
            "files are accepted. Files processed by the dbCAN formatter "
            "`hmmscan-parser.sh` are also accepted using the `dbcan` option."
        )
    )

    parser.add_argument(
        "-n", "--nomenclatures",
        default=None,
        type=argparse.FileType('r'),
        help=(
            "A json formatted file containing the keys nomenclature{1,2,3} "
            "and with values as a list of the classes in the order to be "
            "presented. By default use file in catas/data folder."
        )
    )

    parser.add_argument(
        "-l", "--label",
        dest="labels",
        default=None,
        nargs="+",
        help=(
            "Label to give the prediction for the input file. Specify more "
            "than one label by separating them with a space. The number of "
            "labels should be the same as the number of input files. "
            "By default, the filenames are used as labels."
        )
    )

    parser.add_argument(
        "-o", "--outfile",
        dest="outhandle",
        default=sys.stdout.buffer,
        type=argparse.FileType('wb'),
        help=(
            "File path to write the model to. Default is STDOUT."
        )
    )

    parser.add_argument(
        '--version',
        action='version',
        version='%(prog)s {}'.format(__version__),
        help="Print the version of %(prog)s and exit"
    )

    return parser.parse_args()



#!/usr/bin/env python3

import sys
import traceback
from os import makedirs
from os.path import isfile, basename, splitext, getsize
from os.path import join as pjoin

from typing import List, Optional

from requests.exceptions import HTTPError

from catas.parsers import ParseError

from catas.external import call_hmmscan, call_hmmpress

from catas.cli import MyArgumentError
from catas.pipeline.cli import cli

from catas.count import HMMError
from catas.data import Version, model_filepath

from catas.model import Model

from catas.parsers import FileType
from catas.sanitise import sanitise_fasta, FastaError

from catas.main import runner as catas_runner

from catas import __email__
from catas import __version__
from catas.exitcodes import (
    EXIT_KEYBOARD, EXIT_UNKNOWN, EXIT_INPUT_FORMAT,
    EXIT_IOERR, EXIT_SYSERR
)


def download_dbcan(version: Version, outdir: str) -> str:
    from ..dbcan import DBCAN_URLS, download_file
    from urllib.parse import urlparse
    import os.path

    makedirs(pjoin(outdir, "downloads"), exist_ok=True)
    url = DBCAN_URLS[version]
    url_fname = basename(urlparse(url).path)
    url_fname = pjoin(outdir, "downloads", url_fname)

    if not os.path.exists(url_fname):
        try:
            download_file(url, url_fname)
        except Exception as e:
            if os.path.isfile(url_fname):
                os.remove(url_fname)
                raise e
    return url_fname


def check_fastas(infiles: List[str], correct: bool, outdir: str) -> List[str]:
    from Bio import SeqIO
    errors = []

    if correct:
        new_infiles = []
    else:
        new_infiles = infiles

    if correct:
        makedirs(pjoin(outdir, "fastas"), exist_ok=True)

    for f in infiles:
        with open(f, "r") as handle:
            try:
                seqs = sanitise_fasta(handle, correct)
            except FastaError as e:
                errors.extend(e.messages)
                continue

        if correct:
            new_filename = pjoin(outdir, "fastas", basename(f))
            SeqIO.write(seqs, new_filename, "fasta")
            new_infiles.append(new_filename)

    if len(errors) > 0:
        raise FastaError(errors)

    assert len(infiles) == len(new_infiles)
    return new_infiles


def runner(  # noqa
    infiles: List[str],
    version: Version,
    outdir: str,
    hmms: Optional[str],
    hmmscan_path: str,
    hmmpress_path: str,
    ncpu: int,
    correct: bool,
    quiet: bool,
):
    """ Runs the pipeline. """
    from joblib import Parallel, delayed

    if not quiet:
        print(f"# CATAStrophy v{__version__}\n")

    makedirs(outdir, exist_ok=True)

    if not quiet:
        print("checking Fasta files")
    infiles = check_fastas(infiles, correct, outdir)

    if hmms is None:
        if not quiet:
            print(f"downloading dbCAN {version}")
        hmms = download_dbcan(version, outdir)

    if not quiet:
        print("pressing dbCAN HMMER database")

    call_hmmpress(hmms, cmd=hmmpress_path)

    def parallel_runner(f):
        bname = pjoin(outdir, "search", splitext(basename(f))[0])
        domtab = f"{bname}_domtab.tsv"
        txt = f"{bname}_hmmer.txt"

        # Skip computation
        # The txt file is written after computation so
        # its a better indicator of whether the process succeeded or not
        if isfile(txt) and (getsize(txt) > 0):
            if not quiet:
                print(f"- CACHED: {f}")

            return domtab

        call_hmmscan(f, hmms, domtab, txt, hmmscan_path)

        if not quiet:
            print(f"- COMPLETED: {f}")
        return domtab

    if not quiet:
        print("running hmmscan on proteomes:")

    makedirs(pjoin(outdir, "search"), exist_ok=True)
    domtabs = Parallel(n_jobs=ncpu, prefer="threads")(
        delayed(parallel_runner)(f) for f in infiles
    )

    if not quiet:
        print("\nclassifying proteomes")

    with open(model_filepath(version), "rb") as handle:
        model = Model.read(handle)

    labels = [basename(f[:-len("_domtab.tsv")]) for f in domtabs]

    domtab_handles = [open(h, "r") for h in domtabs]
    rcd_handle = open(pjoin(outdir, "classifications.tsv"), 'w')
    pca_handle = open(pjoin(outdir, "pca.tsv"), 'w')

    try:
        catas_runner(
            domtab_handles,
            rcd_handle,
            pca_handle,
            pjoin(outdir, "counts.tsv"),
            labels,
            FileType.hmmer_domtab,
            model
        )
    except Exception as e:
        for f in domtab_handles:
            f.close()

        rcd_handle.close()
        pca_handle.close()
        raise e

    if not quiet:
        print("Finished!")

    return


def main():  # noqa
    try:
        args = cli(prog=sys.argv[0], args=sys.argv[1:])
    except MyArgumentError as e:
        print(e.message, file=sys.stderr)
        sys.exit(e.errno)

    if args.outdir is None:
        outdir = f"test_{args.version}"
    else:
        outdir = args.outdir

    try:
        runner(
            args.infiles,
            args.model_version,
            outdir,
            args.hmms,
            args.hmmscan_path,
            args.hmmpress_path,
            args.ncpu,
            args.correct,
            args.quiet
        )

    except ParseError as e:
        if e.line is not None:
            header = "Failed to parse file <{}> at line {}.\n".format(
                e.filename, e.line)
        else:
            header = "Failed to parse file <{}>.\n".format(e.filename)

        print("{}\n{}".format(header, e.message), file=sys.stderr)
        sys.exit(EXIT_INPUT_FORMAT)

    except HMMError as e:
        msg = (
            "Encountered an hmm that wasn't present in the training data.\n"
            f"Offending HMMs were: {', '.join(e.hmms)}"
        )
        print(msg, file=sys.stderr)
        sys.exit(EXIT_INPUT_FORMAT)

    except FastaError as e:
        msg = [
            "Encountered an error while checking input fasta files."
        ]
        msg.extend(e.messages)
        print("\n".join(msg), file=sys.stderr)
        sys.exit(EXIT_INPUT_FORMAT)

    except HTTPError as e:

        msg = [
            "Encountered an exception while trying to download the dbCAN "
            "database."
        ]

        if e.response is not None:
            msg.append(f"URL: {e.response.url}")
            msg.append(f"Code: {e.status_code}")

        print("\n".join(msg), file=sys.stderr)
        sys.exit(EXIT_IOERR)

    # except HTTPConnectionPool:
    #     msg = (
    #         "Encountered an exception while trying to download the dbCAN "
    #         "database. It's likely that the URL is broken. \n"
    #         "Please try downloading the database yourself and providing the "
    #         "file."
    #     )
    #     print(msg, file=sys.stderr)
    #     sys.exit(EXIT_IOERR)

    except OSError as e:
        msg = (
            "Encountered a system error.\n"
            "We can't control these, and they're usually related to your OS.\n"
            "Try running again.\n"
        )
        print(msg, file=sys.stderr)
        print(e, file=sys.stderr)
        sys.exit(EXIT_SYSERR)

    except MemoryError:
        msg = (
            "Ran out of memory!\n"
            "Catastrophy shouldn't use much RAM, so check other "
            "processes and try running again."
        )
        print(msg, file=sys.stderr)
        sys.exit(EXIT_SYSERR)

    except KeyboardInterrupt:
        print("Received keyboard interrupt. Exiting.", file=sys.stderr)
        sys.exit(EXIT_KEYBOARD)

    except Exception as e:
        msg = (
            "I'm so sorry, but we've encountered an unexpected error.\n"
            "This shouldn't happen, so please file a bug report with the "
            "authors.\nWe will be extremely grateful!\n\n"
            "You can email us at {}.\n"
            "Alternatively, you can file the issue directly on the repo "
            "<https://github.com/ccdmb/catastrophy/issues>\n\n"
            "Please attach a copy of the following message:"
        ).format(__email__)
        print(e, file=sys.stderr)
        traceback.print_exc(file=sys.stderr)
        sys.exit(EXIT_UNKNOWN)
    return



import argparse

from typing import List

# import datetime

from catas.cli import MyArgumentParser
from catas.data import sample_fasta

# TODAY = datetime.datetime.utcnow().strftime("%Y%m%d")
FASTA_FILE = sample_fasta()


def cli(prog: str, args: List[str]) -> argparse.Namespace:

    parser = MyArgumentParser(description="Generate test data.")

    parser.add_argument(
        "version",
        type=str,
        help=(
            "The dbcan version name"
        )
    )

    parser.add_argument(
        "hmms",
        type=str,
        help=(
            "path to the hmmer formatted database. "
        )
    )

    parser.add_argument(
        "model",
        type=str,
        help=(
            "Path to the model to generate data for."
        )
    )

    # parser.add_argument(
    #     "-t", "--today",
    #     type=str,
    #     default=TODAY,
    #     help="The date to use for the filename."
    # )

    parser.add_argument(
        "-o", "--outdir",
        type=str,
        default=None,
        help="A directory to put the files in. default will be 'test_{version}'"
    )

    parser.add_argument(
        "-i", "--infile",
        type=str,
        default=FASTA_FILE,
        help=(
            "Sample fasta file"
        )
    )

    parser.add_argument(
        "--hmmscan_path",
        type=str,
        default="hmmscan",
        help=(
            "how to look for hmmscan"
        )
    )

    return parser.parse_args()



#!/usr/bin/env python3

import sys
import traceback
from os import makedirs
from os.path import basename, splitext
from os.path import join as pjoin

from catas import parsers
from catas.parsers import FileType
from catas.parsers import ParseError

from catas.cli import MyArgumentError
from catas.gentest.cli import cli

from catas.count import cazy_counts_multi
from catas.count import HMMError

from catas.model import Model, RCDResult, PCAWithLabels

from catas.external import call_hmmscan, call_hmmscan_parser

from catas import __email__
from catas.exitcodes import (
    EXIT_KEYBOARD, EXIT_UNKNOWN, EXIT_INPUT_FORMAT, EXIT_SYSERR
)


def runner(
    infile: str,
    hmms: str,
    model_fname: str,
    version: str,
    outdir: str,
    hmmscan_cmd: str,
):
    """ Runs the pipeline. """

    makedirs(outdir, exist_ok=False)

    # Just touch it to make easier to integrate into package.
    with open(pjoin(outdir, "__init__.py"), "w") as handle:
        print('', file=handle)

    domtab_filename = pjoin(outdir, "hmmer_domtab.tsv")
    hmmer_filename = pjoin(outdir, "hmmer_text.txt")
    dbcan_filename = pjoin(outdir, "hmmer_dbcan.tsv")
    counts_filename = pjoin(outdir, "counts.tsv")
    pca_filename = pjoin(outdir, "pca.tsv")
    rcd_filename = pjoin(outdir, "rcd.tsv")

    with open(model_fname, "rb") as model_handle:
        model = Model.read(model_handle)

    call_hmmscan(infile, hmms, domtab_filename, hmmer_filename, hmmscan_cmd)
    call_hmmscan_parser(domtab_filename, dbcan_filename)

    required_cols = list(model.hmm_lengths.keys())
    with open(dbcan_filename, "r") as dbcan_handle:
        parsed = parsers.parse(
            dbcan_handle,
            format=FileType.dbcan,
            hmm_lens=model.hmm_lengths
        )

        label = splitext(basename(infile))[0]
        counts = cazy_counts_multi([parsed], [label], required_cols)
        counts.write_tsv(counts_filename)

    predictions = model.predict(counts)

    with open(pca_filename, "w") as pca_handle:
        (PCAWithLabels
         .concat([predictions])
         .write_tsv(pca_handle))

    with open(rcd_filename, "w") as rcd_handle:
        RCDResult.write_tsv(rcd_handle, predictions.rcd)

    return


def main():  # noqa
    try:
        args = cli(prog=sys.argv[0], args=sys.argv[1:])
    except MyArgumentError as e:
        print(e.message, file=sys.stderr)
        sys.exit(e.errno)

    if args.outdir is None:
        outdir = f"test_{args.version}"
    else:
        outdir = args.outdir

    try:
        runner(
            args.infile,
            args.hmms,
            args.model,
            args.version,
            outdir,
            args.hmmscan_path
        )

    except ParseError as e:
        if e.line is not None:
            header = "Failed to parse file <{}> at line {}.\n".format(
                e.filename, e.line)
        else:
            header = "Failed to parse file <{}>.\n".format(e.filename)

        print("{}\n{}".format(header, e.message), file=sys.stderr)
        sys.exit(EXIT_INPUT_FORMAT)

    except HMMError as e:
        msg = (
            "Encountered an hmm that wasn't present in the training data.\n"
            f"Offending HMMs were: {', '.join(e.hmms)}"
        )
        print(msg, file=sys.stderr)
        sys.exit(EXIT_INPUT_FORMAT)
        pass

    except OSError as e:
        msg = (
            "Encountered a system error.\n"
            "We can't control these, and they're usually related to your OS.\n"
            "Try running again.\n"
        )
        print(msg, file=sys.stderr)
        print(e.strerror, file=sys.stderr)
        sys.exit(EXIT_SYSERR)

    except MemoryError:
        msg = (
            "Ran out of memory!\n"
            "Catastrophy shouldn't use much RAM, so check other "
            "processes and try running again."
        )
        print(msg, file=sys.stderr)
        sys.exit(EXIT_SYSERR)

    except KeyboardInterrupt:
        print("Received keyboard interrupt. Exiting.", file=sys.stderr)
        sys.exit(EXIT_KEYBOARD)

    except Exception as e:
        msg = (
            "I'm so sorry, but we've encountered an unexpected error.\n"
            "This shouldn't happen, so please file a bug report with the "
            "authors.\nWe will be extremely grateful!\n\n"
            "You can email us at {}.\n"
            "Alternatively, you can file the issue directly on the repo "
            "<https://github.com/ccdmb/catastrophy/issues>\n\n"
            "Please attach a copy of the following message:"
        ).format(__email__)
        print(e, file=sys.stderr)
        traceback.print_exc(file=sys.stderr)
        sys.exit(EXIT_UNKNOWN)
    return



"""
Stored data for use during runtime or testing.

Because dbCAN is an evolving database, we'll have to maintain several models
for each database release. The latest version will always we the default one.
"""

import json
from enum import Enum

from typing import Dict
from typing import List
from typing import Union
from typing import TypeVar, Type

T = TypeVar('T', bound="MyEnum")


def resource_filename(module, resource):
    """ Emulates the behaviour of the old setuptools resource_filename command.

    Basically it just gets rid of the context manager thing, because it's not
    needed. None of the files are zip files or create any temporary files
    that need to be cleaned up.

    This function would be unsafe to use with anything that will be extracted.
    """

    from importlib.resources import path
    with path(module, resource) as handler:
        filename = str(handler)

    return filename


class MyEnum(Enum):
    """ Base class for enums. """

    def __str__(self) -> str:
        return self.name

    @classmethod
    def from_string(cls: Type[T], s: str) -> T:
        try:
            return cls[s]
        except KeyError:
            raise ValueError

    @classmethod
    def from_other(cls: Type[T], f: Union[str, int, "MyEnum"]) -> T:
        if isinstance(f, cls):
            return f
        elif isinstance(f, str):
            return cls.from_string(f)
        elif isinstance(f, int):
            return cls(f)
        else:
            raise ValueError("Expected an enum, string or integer.")


class Version(MyEnum):
    """ Supported DBCAN database versions. """

    v4 = 4
    v5 = 5
    v6 = 6
    v7 = 7
    v8 = 8
    v9 = 9
    v10 = 10

    @classmethod
    def latest(cls) -> "Version":
        return cls.v10


def hmmscan_parser() -> str:
    return resource_filename(__name__, "hmmscan-parser.sh")


def model_filepath(
    version: Union[str, int, Version] = Version.latest()
) -> str:
    """ Loads PCA parameters. """

    # Handle strings as input instead of just enums.
    versionp = Version.from_other(version)

    version_files = {
        Version.v4: resource_filename(__name__, "model_v4_20211001.npz"),
        Version.v5: resource_filename(__name__, "model_v5_20211001.npz"),
        Version.v6: resource_filename(__name__, "model_v6_20211001.npz"),
        Version.v7: resource_filename(__name__, "model_v7_20211001.npz"),
        Version.v8: resource_filename(__name__, "model_v8_20211001.npz"),
        Version.v9: resource_filename(__name__, "model_v9_20211001.npz"),
        Version.v10: resource_filename(__name__, "model_v10_20211001.npz"),
    }

    return version_files[versionp]


def nomenclatures_filepath() -> str:
    """ Loads list containing names and correct order to display classes. """

    return resource_filename(__name__, "nomenclatures.json")


def nomenclatures() -> Dict[str, List[str]]:
    """ Loads list containing names and correct order to display classes. """

    filename = nomenclatures_filepath()

    with open(filename, "r") as handle:
        d = json.load(handle)

    assert isinstance(d, dict)

    for k, v in d.items():
        assert isinstance(v, list)
        assert all(isinstance(vi, str) for vi in v)

    return d


def sample_fasta() -> str:
    # There's no actual version for this.
    return resource_filename(__name__, "test_data.fasta")


def test_files(
    version: Union[str, int, Version] = Version.latest()
) -> Dict[str, str]:
    # Handle strings as input instead of just enums.
    versionp = Version.from_other(version)

    dirs = {
        Version.v4: "test_v4",
        Version.v5: "test_v5",
        Version.v6: "test_v6",
        Version.v7: "test_v7",
        Version.v8: "test_v8",
        Version.v9: "test_v9",
        Version.v10: "test_v10",
    }

    want_files = {
        "hmmer_text": "hmmer_text.txt",
        "hmmer_domtab": "hmmer_domtab.tsv",
        "hmmer_dbcan": "hmmer_dbcan.tsv",
        "pca": "pca.tsv",
        "rcd": "rcd.tsv",
        "counts": "counts.tsv",
    }

    files = dict()

    for key, val in want_files.items():
        files[key] = resource_filename(__name__ + "." + dirs[versionp], val)

    return files



import sys
import traceback

from typing import Sequence
from typing import TextIO
from typing import Optional

from catas.count import cazy_counts_multi, HMMError

from catas import parsers
from catas.parsers import FileType, ParseError
from catas.model import Model, RCDResult, PCAWithLabels
from catas import data
from catas.cli import cli, MyArgumentError

from catas import __email__
from catas.exitcodes import (
    EXIT_KEYBOARD, EXIT_UNKNOWN, EXIT_CLI, EXIT_INPUT_FORMAT, EXIT_SYSERR
)


def runner(
    inhandles: Sequence[TextIO],
    outhandle: TextIO,
    pca_handle: Optional[TextIO],
    counts_filepath: Optional[str],
    labels: Sequence[str],
    file_format: FileType,
    model: Model,
):
    """ Runs the pipeline. """

    parsed = [
        parsers.parse(h, file_format, model.hmm_lengths)
        for h
        in inhandles
    ]
    required_cols = list(model.hmm_lengths.keys())

    counts = cazy_counts_multi(parsed, labels, required_cols)

    if counts_filepath is not None:
        counts.write_tsv(counts_filepath)

    predictions = model.predict(counts)
    RCDResult.write_tsv(outhandle, predictions.rcd)

    if pca_handle is not None:
        (PCAWithLabels
         .concat([model.training_data, predictions])
         .write_tsv(pca_handle))

    return


def main():  # noqa
    """ The cli interface to CATAStrophy. """

    try:
        args = cli(prog=sys.argv[0], args=sys.argv[1:])
    except MyArgumentError as e:
        print(e.message, file=sys.stderr)
        sys.exit(e.errno)

    infile_names = [f.name for f in args.infile]

    if args.labels is None:
        labels = infile_names
    elif len(args.labels) != len(args.infile):
        msg = (
            "argument labels and inhandles: \n"
            "When specified, the number of labels must be the same as the "
            "number of input files. Exiting.\n"
        )

        print(msg, file=sys.stderr)
        sys.exit(EXIT_CLI)
    else:
        labels = args.labels

    if args.model_file is not None:
        model = Model.read(args.model_file)
    else:
        with open(data.model_filepath(args.model_version), "rb") as handle:
            model = Model.read(handle)

    try:
        runner(
            args.infile,
            args.outhandle,
            args.pca_handle,
            args.counts_filepath,
            labels,
            args.file_format,
            model,
        )

    except ParseError as e:
        if e.line is not None:
            header = "Failed to parse file <{}> at line {}.\n".format(
                e.filename, e.line)
        else:
            header = "Failed to parse file <{}>.\n".format(e.filename)

        print("{}\n{}".format(header, e.message), file=sys.stderr)
        sys.exit(EXIT_INPUT_FORMAT)

    except HMMError as e:
        msg = (
            "Encountered an hmm that wasn't present in the training data.\n"
            f"Offending HMMs were: {', '.join(e.hmms)}"
        )
        print(msg, file=sys.stderr)
        sys.exit(EXIT_INPUT_FORMAT)
        pass

    except OSError as e:
        msg = (
            "Encountered a system error.\n"
            "We can't control these, and they're usually related to your OS.\n"
            "Try running again.\n"
        )
        print(msg, file=sys.stderr)
        print(e.strerror, file=sys.stderr)
        sys.exit(EXIT_SYSERR)

    except MemoryError:
        msg = (
            "Ran out of memory!\n"
            "Catastrophy shouldn't use much RAM, so check other "
            "processes and try running again."
        )
        print(msg, file=sys.stderr)
        sys.exit(EXIT_SYSERR)

    except KeyboardInterrupt:
        print("Received keyboard interrupt. Exiting.", file=sys.stderr)
        sys.exit(EXIT_KEYBOARD)

    except Exception as e:
        msg = (
            "I'm so sorry, but we've encountered an unexpected error.\n"
            "This shouldn't happen, so please file a bug report with the "
            "authors.\nWe will be extremely grateful!\n\n"
            "You can email us at {}.\n"
            "Alternatively, you can file the issue directly on the repo "
            "<https://github.com/ccdmb/catastrophy/issues>\n\n"
            "Please attach a copy of the following message:"
        ).format(__email__)
        print(e, file=sys.stderr)
        traceback.print_exc(file=sys.stderr)
        sys.exit(EXIT_UNKNOWN)

    return


if __name__ == '__main__':
    main()



"""
Functions to count the number of specific classes of CAZymes.
"""

from typing import List, Sequence
from typing import Iterable
from typing import Union
from typing import Dict, Set

from collections import defaultdict
import numpy as np

from catas.matrix import Matrix

# These are only used for type checking
from catas.parsers import HMMER, DBCAN


class HMMError(Exception):

    """ Requested an HMM that doesn't exist in this version of dbCAN. """

    def __init__(self, hmms: Sequence[str]):
        """

        Keyword arguments:
        hmms -- a list of hmms that were invalid.
        """

        self.hmms: List[str] = list(hmms)
        return


def cazy_counts_multi(
    handles: Iterable[Union[Iterable[HMMER], Iterable[DBCAN]]],
    labels: Sequence[str],
    required_cols: Sequence[str],
) -> Matrix:

    """ Computes counts for many files. """

    counts = [cazy_counts(m, required_cols) for m in handles]

    return Matrix(
        rows=list(labels),
        columns=list(required_cols),
        arr=np.row_stack(counts)
    )


def cazy_counts(
    matches: Union[Iterable[HMMER], Iterable[DBCAN]],
    required_cols: Sequence[str]
) -> np.ndarray:
    """ Takes a file handle and counts unique occurrences of HMMs.

    Keyword arguments:
    matches -- A list or generator of DBCAN or HMMER records (required).
    required_cols -- A list of HMMs that must have counts in the output.

    Output:
    np.ndarray -- An array of the counts in the same order as the
      required_cols.
    """

    # Loop through all lines and add seqids a set for each HMM that they match.
    these_hmms: Dict[str, Set[str]] = defaultdict(set)
    for match in matches:
        these_hmms[match.hmm].add(match.seqid)

    invalid_hmms = [k for k in these_hmms.keys() if k not in required_cols]
    if len(invalid_hmms) > 0:
        raise HMMError(invalid_hmms)

    row = np.zeros(len(required_cols), dtype=int)

    # Loop through columns and add the required hmm counts to output row.
    for i, col in enumerate(required_cols):
        row[i] = len(these_hmms[col])

    return row



from catas.data import Version

DBCAN_URLS = {
    Version.v4: "http://bcb.unl.edu/dbCAN2/download/Databases/dbCAN-old@UGA/dbCAN-fam-HMMs.txt.v4",  # noqa
    Version.v5: "http://bcb.unl.edu/dbCAN2/download/Databases/dbCAN-old@UGA/dbCAN-fam-HMMs.txt.v5",  # noqa
    Version.v6: "http://bcb.unl.edu/dbCAN2/download/Databases/dbCAN-HMMdb-V6.txt",  # noqa
    Version.v7: "http://bcb.unl.edu/dbCAN2/download/Databases/dbCAN-HMMdb-V7.txt",  # noqa
    Version.v8: "http://bcb.unl.edu/dbCAN2/download/Databases/dbCAN-HMMdb-V8.txt",  # noqa
    Version.v9: "https://bcb.unl.edu/dbCAN2/download/dbCAN-HMMdb-V9.txt",
    Version.v10: "https://bcb.unl.edu/dbCAN2/download/dbCAN-HMMdb-V10.txt",
}


def download_file(url: str, destination: str) -> None:
    import requests
    response = requests.get(url)
    response.raise_for_status()

    with open(destination, "wb") as handle:
        for chunk in response.iter_content(chunk_size=256):
            handle.write(chunk)

    return



from typing import Optional


def call_hmmscan(
    infile: str,
    db: str,
    domtab: str,
    hmmer: str,
    cmd: str = "hmmscan"
) -> None:
    from subprocess import Popen
    from subprocess import PIPE

    command = [
        cmd,
        "--domtblout", domtab,
        db,
        infile
    ]
    call = Popen(command, stdout=PIPE)
    stdout, stderr = call.communicate()

    with open(hmmer, "wb") as handle:
        handle.write(stdout)
    return


def call_hmmpress(
    db: str,
    cmd: str = "hmmpress"
) -> None:
    import os
    from subprocess import Popen

    extensions = [".h3f", ".h3i", ".h3m", ".h3p"]
    all_present = True
    any_present = False

    for ext in extensions:
        fname = f"{db}{ext}"
        if os.path.isfile(fname):
            any_present = True
        else:
            all_present = False

    if all_present:
        return

    if any_present:
        for ext in extensions:
            fname = f"{db}{ext}"
            if os.path.isfile(fname):
                os.remove(fname)

    command = [
        cmd,
        db,
    ]

    call = Popen(command)
    stdout, stderr = call.communicate()
    return


def call_hmmscan_parser(
    infile: str,
    outfile: str,
    cmd: Optional[str] = None
) -> None:
    from subprocess import Popen
    from subprocess import PIPE

    if cmd is None:
        from catas.data import hmmscan_parser
        cmd = hmmscan_parser()

    command = [
        "bash",
        cmd,
        infile
    ]
    call = Popen(command, stdout=PIPE)
    stdout, stderr = call.communicate()

    with open(outfile, "wb") as handle:
        handle.write(stdout)
    return



import argparse
import sys

from typing import List

from catas.parsers import FileType
from catas.data import Version

from catas import __version__

from catas.exitcodes import (
    EXIT_VALID, EXIT_KEYBOARD, EXIT_UNKNOWN, EXIT_CLI, EXIT_INPUT_FORMAT,
    EXIT_INPUT_NOT_FOUND, EXIT_SYSERR, EXIT_CANT_OUTPUT
)


class MyArgumentParser(argparse.ArgumentParser):

    def error(self, message: str):
        """ Override default to have more informative exit codes. """
        self.print_usage(sys.stderr)
        raise MyArgumentError("{}: error: {}".format(self.prog, message))


class MyArgumentError(Exception):

    def __init__(self, message: str):
        self.message = message
        self.errno = EXIT_CLI

        # This is a bit hacky, but I can't figure out another way to do it.
        if "No such file or directory" in message:
            if "infile" in message:
                self.errno = EXIT_INPUT_NOT_FOUND
            elif "outfile" in message:
                self.errno = EXIT_CANT_OUTPUT
        return


def cli(prog: str, args: List[str]) -> argparse.Namespace:

    parser = MyArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=(
            "Examples:\n\n"
            "```bash\n"
            "$ %(prog)s -o prediction.csv dbcan_results.csv\n"
            "$ %(prog)s -l result1 result2 -o prediction.csv "
            "dbcan_results_2.csv dbcan_results_1.csv\n"
            "```\n"
        ),
        epilog=(
            "Exit codes:\n\n"
            f"{EXIT_VALID} - Everything's fine\n"
            f"{EXIT_KEYBOARD} - Keyboard interrupt\n"
            f"{EXIT_CLI} - Invalid command line usage\n"
            f"{EXIT_INPUT_FORMAT} - Input format error\n"
            f"{EXIT_INPUT_NOT_FOUND} - Cannot open the input\n"
            f"{EXIT_SYSERR} - System error\n"
            f"{EXIT_CANT_OUTPUT} - Can't create output file\n"
            f"{EXIT_UNKNOWN} - Unhandled exception, please file a bug!\n"
        )
    )

    parser.add_argument(
        "infile",
        nargs="+",
        type=argparse.FileType('r'),
        help=(
            "Path to the input file output by HMMER or formatted by dbCAN "
            "`hmmscan-parser.sh`. You can specify more than one file by "
            "separating them with a space."
        )
    )

    parser.add_argument(
        "-f", "--format",
        dest="file_format",
        type=FileType.from_string,
        default=FileType.hmmer_domtab,
        choices=list(FileType),
        help=(
            "The format that the input is provided in. If multiple files are "
            "specified, all input must be in the same format. HMMER raw "
            "(hmmer_text) and domain table (hmmer_domtab, default) formatted "
            "files are accepted. Files processed by the dbCAN formatter "
            "`hmmscan-parser.sh` are also accepted using the `dbcan` option."
        )
    )

    parser.add_argument(
        "-l", "--label",
        dest="labels",
        default=None,
        nargs="+",
        help=(
            "Label to give the prediction for the input file. Specify more "
            "than one label by separating them with a space. The number of "
            "labels should be the same as the number of input files. "
            "By default, the filenames are used as labels."
        )
    )

    parser.add_argument(
        "-o", "--outfile",
        dest="outhandle",
        default=sys.stdout,
        type=argparse.FileType('w'),
        help=(
            "File path to write tab delimited output to. Default is STDOUT."
        )
    )

    parser.add_argument(
        "-c", "--counts",
        dest="counts_filepath",
        default=None,
        type=str,
        help="Write the counts table to a file.",
    )

    parser.add_argument(
        "-p", "--pca",
        dest="pca_handle",
        default=None,
        type=argparse.FileType('w'),
        help=(
            "Write the PCA output to a tsv file in addition to the RCDs. "
            "This is useful for plotting the PCAs. "
            "This output includes the training data."
        )
    )

    parser.add_argument(
        "-m", "--model",
        dest="model_version",
        default=Version.latest(),
        type=Version.from_string,
        choices=list(Version),
        help=(
            "The version of the model to use. If you're using old dbCAN "
            "predictions you may have to specify this. The version numbers "
            "are just the versions of dbCAN used to train the models so just "
            "select the dbCAN version that you used. The latest version is "
            "used by default."
        )
    )

    parser.add_argument(
        "--model-file",
        dest="model_file",
        default=None,
        type=argparse.FileType("rb"),
        help=(
            "INTENDED FOR DEVELOPERS. "
            "Instead of the packaged model versions, use this file instead."
        )
    )

    parser.add_argument(
        '--version',
        action='version',
        version='%(prog)s {}'.format(__version__),
        help="Print the version of %(prog)s and exit"
    )

    return parser.parse_args()



""" Utility functions to make programming easier. """

import logging

import functools
import inspect


def log(logger, level=logging.DEBUG):
    """ A decorator function to automatically log function calls.

    This function allows us to track down errors really quickly, as long
    as our functions are reasonably small.

    To use:

    >>> logger = logging.getLogger(__name__)
    >>> @log(logger, logging.DEBUG)
    ... def my_function(one, two):
    ...     return one + two
    >>> my_function(1, 2)
    3

    This would also log to the screen something like...
    DEBUG utils Called my_function(1, 2)
    DEBUG utils Return my_function -> 3

    You can control the level that you want to log these functions at using the
    level parameter, but usually it will just be debug level.
    """

    # Note the extra function is so we can pass arguments to the log decorator.
    def decorator(function):
        # The functools decorator makes it so that the decorator doesn't
        # interfere with the docstrings (or tab autocomplete).
        @functools.wraps(function)
        def wrapper(*args, **kwargs):
            # We get some information about the functions parameters.
            sig = inspect.signature(function).bind(*args, **kwargs)
            sig.apply_defaults()

            # Log the arguments that the function was called with
            logger.log(level, "Called %s%s", function.__name__, sig.args)
            # Call the actual function
            result = function(*args, **kwargs)
            # Log the result of the function
            logger.log(level, "Return %s -> %s", function.__name__, result)
            # Return the result.
            return result
        return wrapper
    return decorator



__program__ = "catastrophy"
__version__ = "0.1.0"
__authors__ = ", ".join(["Darcy Jones", "James Hane"])
__date__ = "31 September 2021"
__email__ = "darcy.a.jones@curtin.edu.au"
__short_blurb__ = (
    "Script to predict lifestyle of filamentous plant pathogens using "
    "carbohydrate-active enzymes (CAZymes)."
)

__license__ = f"""{__program__}-{__version__}

{__short_blurb__}

Copyright (C) {__date__},  {__authors__}

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.
"""



import re
from collections import OrderedDict
import csv

import pathlib  # noqa
from typing import Sequence, Tuple, List
from typing import Mapping, Dict
from typing import NamedTuple, Set
from typing import TextIO, BinaryIO
from typing import Union, Optional, Any

import numpy as np
from numpy.linalg import norm

from catas.matrix import Matrix
from catas import parsers


def groupby_numpy(noms: np.ndarray, m: Matrix, order: Sequence[str]) -> Matrix:
    """ This is a helper funtion to do groupby like things without pandas.

    Keyword arguments:
    noms -- An array containing the classes for each row in the matrix.
    m -- A matrix containing PCs for each sample. Samples as rows.
    order -- A list of nomenclatures in the order that they should
        appear in the output. Note, these must be unique.

    Assertions:
    Each element in `order` must be unique.

    Returns:
    Matrix -- A matrix of centroids for each nomenclature (rows)
        and PC (columns).
    """

    seen: Set[str] = set()
    for o in order:
        assert o not in seen
        seen.add(o)

    centroids = {
        nom: np.mean(m.arr[noms == nom], axis=0)
        for nom
        in np.unique(noms)
    }

    out_arr = np.array([centroids[nom] for nom in order])
    return Matrix(rows=order, columns=m.columns, arr=out_arr)


class CentroidsModel(Matrix):

    @classmethod
    def fit(
        cls,
        m: Matrix,
        classes: Mapping[str, str],
        order: Sequence[str],
    ) -> "CentroidsModel":
        """ Find the centroids given a some labels and a matrix.
        Matrix rownames should match, the keys for the classes typing.

        Keyword arguments:
        m -- A matrix containing PCs (columns) for each label (rows).
        classes -- A dictionary mapping labels to nomenclature classes.
        order -- A list of nomenclatures in the order that they should appear
            in the output.

        Assertions:
        Each element in `order` must be unique.
        """

        noms = np.array([classes[r] for r in m.rows])

        centroids = groupby_numpy(noms, m, order)
        return cls(
            rows=centroids.rows,
            columns=centroids.columns,
            arr=centroids.arr
        )

    def distances(self, points: Matrix) -> Matrix:
        """ Given a point in PCA space, find the distance to each centroid.

        Keyword arguments:
        points -- A Matrix representing the location of sample(s) in PC space.

        Returns:
        Matrix -- A Matrix object giving the euclidean distance to each
            class centroid indexed by the class names.
        """

        # This computes the distance matrix, with samples in rows and
        # centroids as columns.
        diffs = np.apply_along_axis(lambda x: x - self.arr, 1, points.arr)
        normed = np.apply_along_axis(norm, 2, diffs)

        # Grab the class names from the centroids
        new_columns = self.rows

        # return a new series with the class names as index
        return Matrix(rows=points.rows, columns=new_columns, arr=normed)

    @staticmethod
    def _rcd(dists: Matrix) -> Matrix:
        """ Private method to compute rcd.

        This is separate for testing purposes.
        """

        min_ = dists.arr.min(axis=1).reshape(-1, 1)
        max_ = dists.arr.max(axis=1).reshape(-1, 1)
        ratio = (dists.arr - min_) / (max_ - min_)
        rcd_ = 1 - ratio

        # Return the RCDs as a series object with original names/index
        return Matrix(rows=dists.rows, columns=dists.columns, arr=rcd_)

    def rcd(self, points: Matrix) -> Matrix:
        """ Given some points in PCA space, finds the RCD.

        Keyword arguments:
        points -- A Matrix representing the location of sample(s) in PC space.

        Returns:
        Matrix -- A Matrix object with the RCD values for each class.
        """

        dists = self.distances(points)
        return self._rcd(dists)


class NomenclatureClass(NamedTuple):
    """ Contain simple info about training genomes. """

    label: str
    genome: str  # Which species/organism/genome version it is from.
    nomenclature1: str
    nomenclature2: str
    nomenclature3: str

    @classmethod
    def from_tsv(cls, handle: TextIO) -> List["NomenclatureClass"]:
        """ Read from a tsv file.

        The handle must be opened with the newline='' parameter.
        """

        reader = csv.DictReader(handle, delimiter='\t')
        output = list()

        for i, row in enumerate(reader):
            label = row.get("label", None)
            genome = row.get("genome", None)
            n1 = row.get("nomenclature1", None)
            n2 = row.get("nomenclature2", None)
            n3 = row.get("nomenclature3", None)

            if None in (label, genome, n1, n2, n3):
                raise parsers.ParseError(
                    handle.name,
                    i,
                    (
                        "Missing one of the required columns ('label', "
                        "'genome', 'nomenclature1', 'nomenclature2', "
                        "'nomenclature3')."
                    )
                )

            # this is to keep static analysis happy
            assert label is not None
            assert genome is not None
            assert n1 is not None
            assert n2 is not None
            assert n3 is not None

            output.append(cls(label, genome, n1, n2, n3))

        return output

    @classmethod
    def as_numpy_array(
        cls,
        results: Sequence["NomenclatureClass"]
    ) -> np.ndarray:
        """ Convert list of RCD results to a structured array. """

        return np.array(results, dtype=[
            ("label", "<U100"),
            ("genome", "<U100"),
            ("nomenclature1", "<U100"),
            ("nomenclature2", "<U100"),
            ("nomenclature3", "<U100"),
        ])

    @classmethod
    def from_numpy_array(cls, a: np.ndarray) -> List["NomenclatureClass"]:
        return [
            NomenclatureClass(
                str(r["label"]),
                str(r["genome"]),
                str(r["nomenclature1"]),
                str(r["nomenclature2"]),
                str(r["nomenclature3"])
            )
            for r
            in a
        ]


class RCDResult(NamedTuple):
    """ Contain info about RCD values. """

    label: str
    nomenclature: str
    nomenclature_class: str
    value: float

    @classmethod
    def from_matrix(cls, m: Matrix, nomenclature: str) -> List["RCDResult"]:
        """ Convert a wide matrix to a long format list of RCDResults.

        Keyword arguments:
        m -- A matrix to transform from wide to long format.
        nomenclature -- A string containing the nomenclature to repeat in
            output.

        Returns
        A list of RCDResult objects.
        """

        import itertools

        mapped = [
            cls(l, n, c, v)
            for ((l, c), n, v)
            in zip(
                itertools.product(m.rows, m.columns),
                itertools.repeat(nomenclature),
                m.arr.flatten()
            )
        ]
        return mapped

    @classmethod
    def as_numpy_array(cls, results: Sequence["RCDResult"]) -> np.ndarray:
        """ Convert list of RCD results to a structured array. """

        return np.array(results, dtype=[
            ("label", "<U100"),
            ("nomenclature", "<U100"),
            ("nomenclature_class", "<U100"),
            ("value", "float")
        ])

    @classmethod
    def from_numpy_array(cls, a: np.ndarray) -> List["RCDResult"]:
        return [
            RCDResult(
                str(r["label"]),
                str(r["nomenclature"]),
                str(r["nomenclature_class"]),
                float(r["value"])
            )
            for r
            in a
        ]

    @classmethod
    def write_tsv(cls, h: TextIO, a: Sequence["RCDResult"]):
        """ Write a bunch of results to a tsv file.

        Keyword argument:
        h -- A file like object to write to.
        a -- A list of RCDResults as output from `to_predictions_array`.
        """

        sorted_results = sorted(
            a,
            key=lambda r: (r.nomenclature, r.label,
                           -1 * r.value, r.nomenclature_class)
        )

        np.savetxt(
            h,
            RCDResult.as_numpy_array(sorted_results),
            delimiter='\t',
            fmt=["%s", "%s", "%s", "%f"],
            header="label\tnomenclature\tclass\tvalue"
        )
        return


class PCAModel(object):

    def __init__(self, means: np.ndarray, components: np.ndarray):
        """ Stores the necessary data and methods to do PCA transformations.
        """
        self.means = means
        self.components = components
        return

    @classmethod
    def fit(cls, counts: Matrix, n_components: int = 16) -> "PCAModel":
        """ Fit the PC model to find the loadings etc.

        Keyword arguments:
            counts -- A matrix of CAZyme counts.
            n_components -- The number of components to retain in the model.
            """
        # This is not a package dependency because we don't
        # expect regular users to train their own models.
        from sklearn.decomposition import PCA

        pipeline = PCA(n_components=n_components, svd_solver="full")
        pipeline.fit(counts.arr)

        return cls(pipeline.mean_, pipeline.components_)

    def transform(self, counts: Matrix) -> Matrix:
        """ Takes a matrix of CAZyme counts and gets PCA transformed values.

        NB this is based on the sklearn implementation.
        Used to avoid compatibility issues between skl versions.

        Keyword arguments:
        counts -- A Matrix containing CAZyme counts.

        Returns:
        Matrix -- A matrix object with the PCA transformed values.
        """

        values = counts.arr.astype(float)
        X = values - self.means
        X_transformed = np.dot(X, self.components.T)

        columns = [
            "pc{:0>2}".format(i + 1)
            for i
            in range(X_transformed.shape[1])
        ]

        return Matrix(rows=counts.rows, columns=columns, arr=X_transformed)

    def as_dict_of_arrays(self, prefix: str = "") -> Dict[str, np.ndarray]:
        return {
            f"{prefix}means": self.means,
            f"{prefix}components": self.components,
        }

    @classmethod
    def from_dict_of_arrays(
        cls,
        doa: Mapping[str, np.ndarray],
        prefix: str = ""
    ) -> "PCAModel":
        return cls(doa[f"{prefix}means"], doa[f"{prefix}components"])


class HMMLengths(OrderedDict):

    """ A subclass of an ordered dictionary. """

    @classmethod
    def read_hmm(cls, handle: TextIO) -> "HMMLengths":
        """ Find the lengths of the HMMs.

        Keyword arguments:
        handle -- A file-like object to a HMMER formatted text database.
        """

        regex = re.compile(r"\s+")
        output = list()

        current_name = None

        for line in handle:
            if line.startswith("NAME"):
                line = regex.split(line.strip(), maxsplit=1)[1]
                current_name = parsers.split_hmm(line)

            elif line.startswith("LENG"):
                line = regex.split(line.strip(), maxsplit=1)[1]

                assert current_name is not None, line
                output.append((current_name, int(line)))
                current_name = None

        # Sort by the hmm name
        output.sort(key=lambda t: t[0])
        return cls(output)

    def to_array(self) -> np.ndarray:
        """ Convert to a numpy structured array. """
        return np.array(
            list(self.items()),
            dtype=[("hmm", 'U60'), ("count", 'uint64')]
         )

    @classmethod
    def from_array(cls, arr: np.ndarray) -> "HMMLengths":
        """ Create from a numpy structured array. """
        return cls(arr)


class PCAWithLabels(object):

    def __init__(
        self,
        pca: Matrix,
        rcd: Sequence[RCDResult],
        classes: Optional[Sequence[NomenclatureClass]] = None,
    ):
        """ Keep the rcd and PCA results together.

        Return type for the models.
        """

        self.pca = pca
        self.rcd = rcd
        self.classes = classes
        return

    def as_dict_of_arrays(self, prefix: str = "") -> Dict[str, np.ndarray]:
        out = {
            f"{prefix}rcd": RCDResult.as_numpy_array(self.rcd),
        }

        out.update(self.pca.as_dict_of_arrays(f"{prefix}pca_"))

        if self.classes is not None:
            out[f"{prefix}classes"] = NomenclatureClass.as_numpy_array(
                self.classes
            )
        return out

    @classmethod
    def from_dict_of_arrays(
        cls,
        doa: Mapping[str, np.ndarray],
        prefix: str = ""
    ) -> "PCAWithLabels":
        rcd = RCDResult.from_numpy_array(doa[f"{prefix}rcd"])
        pca = Matrix.from_dict_of_arrays(doa, f"{prefix}pca_")

        if f"{prefix}classes" in doa:
            classes: Optional[List[NomenclatureClass]] = (
                NomenclatureClass
                .from_numpy_array(doa[f"{prefix}classes"])
            )
        else:
            classes = None
        return cls(pca, rcd, classes)

    @classmethod
    def concat(cls, pwls: Sequence["PCAWithLabels"]) -> "PCAWithLabels":
        pca = Matrix.concat([p.pca for p in pwls])

        rcd: List[RCDResult] = []
        classes: Optional[List[NomenclatureClass]] = None
        for p in pwls:
            rcd.extend(p.rcd)

            if p.classes is not None:
                if classes is None:
                    classes = []

                classes.extend(p.classes)

        return cls(pca, rcd, classes)

    def write_tsv(self, h: TextIO, threshold: float = 0.8):
        "Write a copy if the results to a tsv file."

        if self.classes is not None:
            classes: Optional[Dict[str, NomenclatureClass]] = {
                r.label: r for r in self.classes
            }
        else:
            classes = None

        best_rcds: Dict[Tuple[str, str], RCDResult] = dict()
        sig_rcds: Dict[Tuple[str, str], List[str]] = dict()

        for rcd in self.rcd:
            id_ = (rcd.nomenclature, rcd.label)
            if rcd.value >= threshold:
                if id_ not in sig_rcds:
                    sig_rcds[id_] = [rcd.nomenclature_class]
                else:
                    sig_rcds[id_].append(rcd.nomenclature_class)

            if id_ not in best_rcds:
                best_rcds[id_] = rcd
            elif rcd.value > best_rcds[id_].value:
                best_rcds[id_] = rcd

        # First we setup the columns.
        columns = [("label", "<U100")]
        if self.classes is not None:
            columns.extend([
                ("genome", "<U100"),
                ("nomenclature1", "<U100"),
                ("nomenclature2", "<U100"),
                ("nomenclature3", "<U100"),
            ])

        columns.extend([
            ("nomenclature1_pred", "<U100"),
            ("nomenclature2_pred", "<U100"),
            ("nomenclature3_pred", "<U100"),
            ("nomenclature1_ancillary", "<U300"),
            ("nomenclature2_ancillary", "<U300"),
            ("nomenclature3_ancillary", "<U300"),
        ])

        columns.extend([(p, "<U10") for p in self.pca.columns])

        # Using any here so we can have length of tuple depend on matrix size.
        rows: List[Any] = list()
        for label, matrix_row in zip(self.pca.rows, self.pca.arr):
            row: List[str] = [label]

            if classes is not None:
                clss = classes.get(label, None)
                if clss is None:
                    row.extend(list("...."))
                else:
                    row.extend([
                        classes[label].genome,
                        classes[label].nomenclature1,
                        classes[label].nomenclature2,
                        classes[label].nomenclature3,
                    ])

            row.extend([
                best_rcds[("nomenclature1", label)].nomenclature_class,
                best_rcds[("nomenclature2", label)].nomenclature_class,
                best_rcds[("nomenclature3", label)].nomenclature_class,
                ",".join(sig_rcds[("nomenclature1", label)]),
                ",".join(sig_rcds[("nomenclature2", label)]),
                ",".join(sig_rcds[("nomenclature3", label)]),
            ])

            row.extend(map(str, list(matrix_row)))

            assert len(row) == len(columns)
            rows.append(tuple(row))

        np.savetxt(
            h,
            np.array(rows, dtype=columns),
            delimiter="\t",
            fmt=["%s" for _ in columns],
            header="\t".join(c[0] for c in columns)
        )
        return


class Model(object):

    def __init__(
        self,
        n1_centroids: CentroidsModel,
        n2_centroids: CentroidsModel,
        n3_centroids: CentroidsModel,
        pca_model: PCAModel,
        hmm_lengths: HMMLengths,
        training_data: PCAWithLabels,
    ):
        """ A Model contains all of the bits and methods to predict RCDS.

        Keyword arguments:
        n1_centroids -- A centroids model containing the centroids for each PC
            for each nomenclature one class.
        n2_centroids -- As for n1_centroids but for nomenclature two.
        n3_centroids -- As for n1_centroids but for nomenclature three.
        """

        self.n1_centroids = n1_centroids
        self.n2_centroids = n2_centroids
        self.n3_centroids = n3_centroids
        self.pca_model = pca_model
        self.hmm_lengths = hmm_lengths
        self.training_data = training_data
        return

    @classmethod
    def fit(
        cls,
        counts: Matrix,
        classes: Sequence[NomenclatureClass],
        nomenclatures: Mapping[str, Sequence[str]],
        hmm_lengths: HMMLengths,
        n_components: int = 16,
    ) -> "Model":
        """ Train the models given some counts.

        Keyword arguments:
        counts -- A matrix of counts for each hmm (columns), and label (rows).
        classes -- A list of nomenclature class objects.
        nomenclatures -- A dict mapping nomenclature versions to a list
            of the nomenclature names in the order that they should
            appear in the output.
        hmm_lengths -- A hmm lengths object mapping hmm names to the length
            of the model.
        n_components -- The number of components to include in the model.
        """

        # Fit the PC models.
        pca = PCAModel.fit(counts, n_components)

        # Get the actual pca transformed values.
        trans = pca.transform(counts)

        # Get a dict mapping labels to nomenclature1 classes.
        n1_dict = {c.label: c.nomenclature1 for c in classes}
        n2_dict = {c.label: c.nomenclature2 for c in classes}
        n3_dict = {c.label: c.nomenclature3 for c in classes}

        # Get a non-redundant list of classes in sorted order.
        n1_classes = sorted(set(n1_dict.values()))

        # Find the centroids
        n1_centroids = CentroidsModel.fit(
            trans,
            n1_dict,
            n1_classes
        )

        n2_centroids = CentroidsModel.fit(
            trans,
            n2_dict,
            sorted(set(n2_dict.values())),
        )

        n3_centroids = CentroidsModel.fit(
            trans,
            n3_dict,
            sorted(set(n3_dict.values())),
        )

        # Get the RCD predictions for the training data.
        n1_rcd = n1_centroids.rcd(trans)
        n2_rcd = n2_centroids.rcd(trans)
        n3_rcd = n3_centroids.rcd(trans)

        rcd = RCDResult.from_matrix(n1_rcd, "nomenclature1")
        rcd.extend(RCDResult.from_matrix(n2_rcd, "nomenclature2"))
        rcd.extend(RCDResult.from_matrix(n3_rcd, "nomenclature3"))

        return cls(
            n1_centroids,
            n2_centroids,
            n3_centroids,
            pca,
            hmm_lengths,
            PCAWithLabels(trans, rcd, classes)
        )

    def predict(self, counts: Matrix) -> PCAWithLabels:
        """ Given a matrix of counts, find the PC transformation and the RCDs.

        Keyword arguments:
        counts -- A `Matrix` containing the cazyme counts (required).

        Returns:
        Array -- A structured array containing RCD values.
        Matrix -- A matrix containing the PC transformed counts.
        """

        pca_transformed = self.pca_model.transform(counts)
        n1_rcd = self.n1_centroids.rcd(pca_transformed)
        n2_rcd = self.n2_centroids.rcd(pca_transformed)
        n3_rcd = self.n3_centroids.rcd(pca_transformed)

        rcd = RCDResult.from_matrix(n1_rcd, "nomenclature1")
        rcd.extend(RCDResult.from_matrix(n2_rcd, "nomenclature2"))
        rcd.extend(RCDResult.from_matrix(n3_rcd, "nomenclature3"))

        return PCAWithLabels(pca_transformed, rcd)

    def as_dict_of_arrays(self, prefix="") -> Dict[str, np.ndarray]:
        """ Convert the object to something that we can save easily. """

        output = {f"{prefix}hmm_lengths": self.hmm_lengths.to_array()}

        output.update(
            self.pca_model.as_dict_of_arrays(prefix=f"{prefix}pca_model_")
        )

        output.update(
            self.n1_centroids.as_dict_of_arrays(
                prefix=f"{prefix}n1_centroids_"
            )
        )
        output.update(
            self.n2_centroids.as_dict_of_arrays(
                prefix=f"{prefix}n2_centroids_"
            )
        )
        output.update(
            self.n3_centroids.as_dict_of_arrays(
                prefix=f"{prefix}n3_centroids_"
            )
        )

        output.update(
            self.training_data.as_dict_of_arrays(
                prefix=f"{prefix}training_data_"
            )
        )

        return output

    @classmethod
    def from_dict_of_arrays(
        cls,
        doa: Mapping[str, np.ndarray],
        prefix: str = ""
    ) -> "Model":
        """ Create new model from saved version.

        Intended to be used on np.load output.
        """

        hmm_lengths = HMMLengths.from_array(doa[f"{prefix}hmm_lengths"])

        pca_model = PCAModel.from_dict_of_arrays(
            doa,
            prefix=f"{prefix}pca_model_"
        )

        n1_centroids = CentroidsModel.from_dict_of_arrays(
            doa,
            prefix=f"{prefix}n1_centroids_"
        )

        n2_centroids = CentroidsModel.from_dict_of_arrays(
            doa,
            prefix=f"{prefix}n2_centroids_"
        )

        n3_centroids = CentroidsModel.from_dict_of_arrays(
            doa,
            prefix=f"{prefix}n3_centroids_"
        )

        training_data = PCAWithLabels.from_dict_of_arrays(
            doa,
            prefix=f"{prefix}training_data_"
        )

        # This is to appease the typechecker.
        assert isinstance(n1_centroids, CentroidsModel)
        assert isinstance(n2_centroids, CentroidsModel)
        assert isinstance(n3_centroids, CentroidsModel)

        assert isinstance(training_data, PCAWithLabels)

        return cls(n1_centroids, n2_centroids, n3_centroids,
                   pca_model, hmm_lengths, training_data)

    def write(
        self,
        handle: Union[str, BinaryIO, "pathlib.Path"],
        prefix: str = ""
    ):
        """ Write a model to a numpy binary file. """

        np.savez(
            handle,
            **self.as_dict_of_arrays(prefix=prefix)
        )
        return

    @classmethod
    def read(
        cls,
        handle: Union[str, BinaryIO, "pathlib.Path"],
        prefix: str = "",
    ) -> "Model":
        """ Read a model from a numpy binary file. """

        f = np.load(handle, allow_pickle=False)
        return cls.from_dict_of_arrays(f, prefix=prefix)



"""
A simpler alternative to a pandas dataframe for non-interactive work.
"""

import csv
import pathlib  # noqa
from typing import cast
from typing import List, Sequence
from typing import Dict, Mapping
from typing import Union
from typing import BinaryIO, TextIO

import numpy as np


class Matrix(object):
    """ A simple wrapper around a numpy array that tracks column and row names.

    Order is row major.
    Using this as an alternative to pandas to get rid of dependency.
    """

    def __init__(
        self,
        rows: Sequence[str],
        columns: Sequence[str],
        arr: np.ndarray,
    ):
        assert len(rows) == arr.shape[0]
        assert len(columns) == arr.shape[1]

        self.rows = list(rows)
        self.columns = list(columns)
        self.arr = arr
        return

    def __getitem__(self, key: Union[str, int]) -> np.ndarray:
        if isinstance(key, str):
            key = self.columns.index(key)

        return self.arr[:, key]

    @classmethod
    def from_row(
        cls,
        row: str,
        columns: Sequence[str],
        arr: np.ndarray
    ) -> "Matrix":
        """ Create a matrix from a single row """

        return cls([row], list(columns), arr.reshape(1, -1))

    @classmethod
    def concat(cls, dfs: Sequence["Matrix"]) -> "Matrix":
        """ Concatenate multiple matrices together by rows (ie. columns are
        matched)
        """

        if len(dfs) == 0:
            return cls([], [], np.zeros((0, 0)))

        columns = dfs[0].columns

        # Columns must all be same shape and names.
        for df in dfs:
            assert df.columns == columns

        # Concatenates and flattens all rows.
        rows = [r for df in dfs for r in df.rows]
        # NB concatenate requires all dfs to have correct dimensions.
        # IE they can't be a flat array.
        arr = np.concatenate([df.arr for df in dfs])
        return cls(rows, columns, arr)

    def as_df(self):
        # This isn't typed so that I don't need pandas as dependency
        import pandas as pd
        return pd.DataFrame(
            data=self.arr,
            index=self.rows,
            columns=self.columns
        )

    @classmethod
    def from_df(cls, df):
        """ Assumes heterogeneous df with rownames. """
        rows = df.index.tolist()
        columns = df.columns.tolist()
        return cls(rows, columns, df.values)

    def as_dict_of_arrays(
        self,
        prefix: str = ""
    ) -> Dict[str, np.ndarray]:
        return {
            f"{prefix}arr": self.arr,
            f"{prefix}rows": np.array(self.rows, dtype='U60'),
            f"{prefix}columns": np.array(self.columns, dtype='U60'),
        }

    @classmethod
    def from_dict_of_arrays(
        cls,
        doa: Mapping[str, np.ndarray],
        prefix: str = ""
    ) -> "Matrix":
        return cls(
            doa[f"{prefix}rows"].tolist(),
            doa[f"{prefix}columns"].tolist(),
            doa[f"{prefix}arr"],
        )

    def write(
        self,
        handle: Union[str, BinaryIO, "pathlib.Path"],
        prefix: str = ""
    ):
        """ Writes matrix out as a numpy file. """
        np.savez(
            handle,
            *self.as_dict_of_arrays(prefix=prefix)
        )
        return

    @classmethod
    def read(
        cls,
        handle: Union[str, BinaryIO, "pathlib.Path"],
        prefix: str = "",
    ) -> "Matrix":
        """ Reads matrix from an numpy file. """
        f = np.load(handle, allow_pickle=False)
        return cls.from_dict_of_arrays(f)

    def write_tsv(
        self,
        filename: Union[str, TextIO],
        rows_colname: str = "# label"
    ):
        """ Write a tsv formatted table from the matrix.

        Note that if you provide a file-object handle to filename, it must
        have been opened with the `newline=''` option (required by csv).
        """
        try:
            if isinstance(filename, str):
                handle: TextIO = open(filename, 'w', newline='')
            else:
                handle = filename

            tab_writer = csv.writer(handle, "excel-tab")

            # Write the header
            tab_writer.writerow([rows_colname] + self.columns)

            # Write the lines
            for rowname, row in zip(self.rows, self.arr):
                tab_writer.writerow([rowname] + list(row))

        finally:
            if isinstance(filename, str):
                handle.close()
        return

    def as_serializable(self) -> List[Dict[str, Union[str, float]]]:
        """ Output the matrix as a dict, which can be easily serialised into
        JSON etc.
        """

        output = list()
        for rowname, row in zip(self.rows, self.arr):
            drow: Dict[str, Union[str, float]] = {
                str(k): float(v)
                for k, v
                in zip(self.columns, row)
            }
            drow["label"] = rowname
            output.append(drow)
        return output

    @classmethod
    def from_serialized(
        cls,
        ser: Sequence[Mapping[str, Union[str, float]]]
    ) -> "Matrix":
        """ Parses a dictionary representation back into a matrix. """

        if len(ser) == 0:
            return cls([], [], np.zeros((0, 0)))

        rows: List[str] = list()
        arr = list()
        columns: List[str] = [k for k, v in ser if k != "label"]
        for row in ser:
            arr_row = list()

            # This helps the type checker.
            # It shouldn't occur at runtime if the spec is observed.
            assert isinstance(row["label"], str)
            rows.append(row["label"])

            for column in columns:
                arr_row.append(row[column])
            arr.append(arr_row)

        return cls(
            rows,
            columns,
            np.array(arr),
        )



import argparse

from typing import List

from catas.cli import MyArgumentParser
from catas.data import Version

from catas import __version__

from catas.exitcodes import (
    EXIT_VALID, EXIT_KEYBOARD, EXIT_UNKNOWN, EXIT_CLI, EXIT_INPUT_FORMAT,
    EXIT_INPUT_NOT_FOUND, EXIT_SYSERR, EXIT_CANT_OUTPUT
)


def cli(prog: str, args: List[str]) -> argparse.Namespace:

    import joblib

    parser = MyArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=(
            "Examples:\n\n"
            "```bash\n"
            "$ %(prog)s -o outdir proteome_1.fasta proteome_2.fasta\n"
            "```\n"
        ),
        epilog=(
            "Exit codes:\n\n"
            f"{EXIT_VALID} - Everything's fine\n"
            f"{EXIT_KEYBOARD} - Keyboard interrupt\n"
            f"{EXIT_CLI} - Invalid command line usage\n"
            f"{EXIT_INPUT_FORMAT} - Input format error\n"
            f"{EXIT_INPUT_NOT_FOUND} - Cannot open the input\n"
            f"{EXIT_SYSERR} - System error\n"
            f"{EXIT_CANT_OUTPUT} - Can't create output file\n"
            f"{EXIT_UNKNOWN} - Unhandled exception, please file a bug!\n"
        )
    )

    parser.add_argument(
        "infiles",
        type=str,
        nargs="+",
        help=(
            "Proteome fasta-files to run."
        )
    )

    parser.add_argument(
        "-m", "--model",
        dest="model_version",
        default=Version.latest(),
        type=Version.from_string,
        choices=list(Version),
        help=(
            "The version of the model to use. If you're using old dbCAN "
            "predictions you may have to specify this. The version numbers "
            "are just the versions of dbCAN used to train the models so just "
            "select the dbCAN version that you used. The latest version is "
            "used by default."
        )
    )

    parser.add_argument(
        "--hmms",
        type=str,
        help=(
            "Path to the dbCAN hmmer-formatted database. "
            "Note that the specified model version must match the version of "
            "the database specified here."
        )
    )

    parser.add_argument(
        "-o", "--outdir",
        type=str,
        default="results",
        help="A directory to put the files in. default will be 'results'"
    )

    parser.add_argument(
        "--hmmscan_path",
        type=str,
        default="hmmscan",
        help=(
            "Where to look for hmmscan"
        )
    )

    parser.add_argument(
        "--hmmpress_path",
        type=str,
        default="hmmpress",
        help=(
            "Where to look for hmmpress"
        )
    )

    parser.add_argument(
        "--ncpu",
        type=int,
        default=joblib.cpu_count(),
        help="How many processes to run."
    )

    parser.add_argument(
        "-c", "--correct",
        action="store_true",
        default=False,
        help="Replace invalid characters in Fasta files."
    )

    parser.add_argument(
        "-q", "--quiet",
        action="store_true",
        default=False,
        help="Suppress running feedback."
    )

    parser.add_argument(
        '--version',
        action='version',
        version='%(prog)s {}'.format(__version__),
        help="Print the version of %(prog)s and exit"
    )

    return parser.parse_args()



"""
"""

import pytest
from numpy.testing import assert_almost_equal

from catas.parsers import HMMER
from catas.count import cazy_counts_multi
from catas.model import Model

from catas.data import test_files
from catas.data import model_filepath
from catas.data import Version


@pytest.mark.parametrize("version,clss,exp_val", [
    (Version.v10, "monomertroph3", 1.0),
    (Version.v10, "saprotroph", 0.8773836040317411),
    (Version.v10, "vasculartroph", 0.0),
    (Version.v10, "mesotroph_intracellular", 0.08630181053724584),
    (Version.v5, "monomertroph3", 0.9676738995420295),
    (Version.v5, "saprotroph", 0.8796622614136591),
    (Version.v5, "vasculartroph", 0.0),
    (Version.v5, "mesotroph_intracellular", 0.21770575123709068),
    (Version.v4, "monomertroph3", 0.9674518271503275),
    (Version.v4, "saprotroph", 0.879052715572339),
    (Version.v4, "vasculartroph", 0.0),
    (Version.v4, "mesotroph_intracellular", 0.21759385711517643),
])
def test_predict(version, clss, exp_val):
    """ Pretty much a repeat of test_rcds. """

    with open(model_filepath(version), "rb") as handle:
        model = Model.read(handle)

    required_cols = list(model.hmm_lengths.keys())

    files = test_files(version)
    with open(files["hmmer_text"], "r") as handle:
        parsed = HMMER.from_file(
            handle,
            model.hmm_lengths,
            "hmmer3-text",
        )
        counts = cazy_counts_multi([parsed], ["test"], required_cols)

    pred = model.predict(counts)
    rcds = pred.rcd

    rcds_dict = dict()
    for res in rcds:
        assert res.label == "test"
        rcds_dict[(res.label, res.nomenclature,
                   res.nomenclature_class)] = res.value

    # Checks that results are accurate to 5 decimal places.
    assert_almost_equal(
        exp_val,
        rcds_dict[("test", "nomenclature3", clss)],
        decimal=5
    )
    return


@pytest.mark.parametrize("version,pc,exp_val", [
    (Version.v4, "pc01", -69.59756347101215),
    (Version.v4, "pc02", -13.422481929003107),
    (Version.v5, "pc01", -69.56184489008383),
    (Version.v5, "pc02", -13.395714644864263),
    (Version.v6, "pc01", -65.58320320620606),
    (Version.v6, "pc02", -11.363653203632293),
])
def test_pca(version, pc, exp_val):

    with open(model_filepath(version), "rb") as handle:
        model = Model.read(handle)

    required_cols = list(model.hmm_lengths.keys())

    files = test_files(version)
    with open(files["hmmer_text"], "r") as handle:
        parsed = HMMER.from_file(
            handle,
            model.hmm_lengths,
            "hmmer3-text",
        )
        counts = cazy_counts_multi([parsed], ["test"], required_cols)

    pred = model.predict(counts)
    trans = pred.pca

    # Checks that results are accurate to 5 decimal places.
    column_index = trans.columns.index(pc)
    assert_almost_equal(exp_val, trans.arr[0, column_index], decimal=5)
    return



import pytest

from catas.data import Version
from catas.data import model_filepath
from catas.data import test_files
from catas.model import Model
from catas.parsers import split_hmm
from catas.parsers import HMMER


@pytest.mark.parametrize("hmm,exp", [
    ("GH80.hmm", "GH80"),
    ("AC1.hmm", "AC1"),
    ("AC1.phmm", "AC1"),
    ("AC1h.hmm", "AC1h")
])
def test_split_hmm(hmm, exp):
    result = split_hmm(hmm)
    assert result == exp
    return


@pytest.mark.parametrize("version,idx,col,exp_val", [
    (Version.v4, 0, "hmm", "GH47"),
    (Version.v4, 1, "hmm_len", 29),
    (Version.v4, 2, "seqid", "tr|Q0UA13|Q0UA13_PHANO"),
    (Version.v4, 6, "hmm", "GH10"),
    (Version.v4, 11, "seqid", "tr|Q68KX9|Q68KX9_PHAND"),
    (Version.v5, 0, "hmm", "GH47"),
    (Version.v5, 2, "seqid", "tr|Q0UA13|Q0UA13_PHANO"),
    (Version.v5, 6, "hmm", "GH10"),
    (Version.v5, 11, "seqid", "tr|Q68KX9|Q68KX9_PHAND"),
    (Version.v6, 0, "hmm", "GH47"),
    (Version.v6, 1, "hmm_len", 29),
    (Version.v6, 2, "seqid", "tr|Q0UA13|Q0UA13_PHANO"),
    (Version.v6, 6, "hmm", "GH10"),
    (Version.v6, 11, "seqid", "tr|Q68KX9|Q68KX9_PHAND"),
    (Version.v7, 0, "hmm", "GH47"),
    (Version.v7, 1, "hmm_len", 29),
    (Version.v7, 2, "seqid", "tr|Q0UA13|Q0UA13_PHANO"),
    (Version.v7, 6, "hmm", "GH10"),
    (Version.v7, 11, "seqid", "tr|Q68KX6|Q68KX6_PHAND"),
    (Version.v8, 0, "hmm", "GH47"),
    (Version.v8, 1, "hmm_len", 29),
    (Version.v8, 2, "seqid", "tr|Q0UA13|Q0UA13_PHANO"),
    (Version.v8, 6, "hmm", "GH10"),
    (Version.v8, 11, "seqid", "tr|Q68KX6|Q68KX6_PHAND"),
])
def test_parse_hmmer_text_output(version, idx, col, exp_val):

    with open(model_filepath(version), "rb") as handle:
        model = Model.read(handle)

    files = test_files(version)
    with open(files["hmmer_text"], "r") as handle:
        sample = list(HMMER.from_file(
            handle,
            model.hmm_lengths,
            "hmmer3-text",
        ))

    assert getattr(sample[idx], col) == exp_val
    return



from io import StringIO
from enum import Enum
import sys
import re

from typing import NamedTuple
from typing import TextIO, BinaryIO
from typing import Union, Optional
from typing import Sequence, List
from typing import Dict
from typing import Iterator

from Bio import SearchIO

# These are only used for type checking
from Bio.SearchIO._model.query import QueryResult
from Bio.SearchIO._model.hit import Hit
from Bio.SearchIO._model.hsp import HSP


REGEX = re.compile(r"\.p?hmm$")
TEXT_FIRSTLINE = re.compile(r"^# hmmscan ::")  # noqa: E501
DOMTAB_FIRSTLINE = re.compile(r"^#\s+--- full sequence ---.*$")  # noqa: E501
DBCAN_FIRSTLINE = re.compile("\t".join([
    r"\S+", "[0-9]+", r"\S+", "[0-9]+",
    r"([0-9]*[.])?[0-9]+([eE][-+]?\d+)?",
    "[0-9]+", "[0-9]+", "[0-9]+", "[0-9]+",
    r"([0-9]*[.])?[0-9]+([eE][-+]?\d+)?",
]))


def predict_filetype(handle: TextIO):

    handle.seek(0)
    line = ""

    # Just in case there is a blank line at the top.
    for line in handle:
        line = line.strip()
        if line != "":
            break

    if line == "":
        msg = (
            "The input file does not contain any non-empty lines.\n"
            "Please check your inputs and if you believe that "
            "this is an error, please contact us."
        )

        if hasattr(handle, "name"):
            name: Optional[str] = handle.name
        else:
            name = None

        raise ParseError(name, None, msg)

    if TEXT_FIRSTLINE.match(line):
        return "hmmer_text"
    elif DOMTAB_FIRSTLINE.match(line):
        return "hmmer_domtab"
    elif DBCAN_FIRSTLINE.match(line):
        return "dbcan"
    else:
        msg = (
            "The input file does not appear to be in any "
            "of the supported formats.\n"
            "Please check your inputs and if you believe that "
            "this is an error, please contact us."
        )

        if hasattr(handle, "name"):
            name = handle.name
        else:
            name = None

        raise ParseError(name, None, msg)


class FileType(Enum):
    """ All possible input filetypes that we support. """

    dbcan = 1
    hmmer_text = 2
    hmmer_domtab = 3

    def __str__(self) -> str:
        return self.name

    @classmethod
    def from_string(cls, s: str) -> "FileType":
        try:
            return cls[s]
        except KeyError:
            raise ValueError


class ParseError(Exception):
    """ Some aspect of parsing failed. """

    def __init__(
        self,
        filename: Optional[str],
        line: Optional[int],
        message: str
    ):
        self.filename = filename
        self.line = line
        self.message = message
        return


class LineParseError(Exception):

    def __init__(self, message: str):
        self.message = message
        return


def parse(
    handle: TextIO,
    format: Union[str, FileType],
    hmm_lens: Optional[Dict[str, int]],
) -> Union[Iterator["HMMER"], Iterator["DBCAN"]]:
    """ Wrapper that directs parsing bases on enum.

    Keyword arguments:
    handle -- A file-like object to parse.
    format -- The file-format that the handle is in, should be a FileType
        object.
    hmm_lens -- Required only for the HMMer formats. A dictionary mapping the
        hmm names to the length of the HMM model.

    Returns:
    A generator over the parsed records.
    Records will be either HMMER or DBCAN objects.
    """

    if isinstance(format, str):
        format = FileType[format]

    if format == FileType.dbcan:
        return DBCAN.from_file(handle)

    elif format == FileType.hmmer_domtab:
        assert hmm_lens is not None
        parsed = HMMER.from_file(
            handle,
            format="hmmscan3-domtab",
            hmm_lens=hmm_lens,
        )
        return parsed

    elif format == FileType.hmmer_text:
        assert hmm_lens is not None
        return HMMER.from_file(
            handle,
            format="hmmer3-text",
            hmm_lens=hmm_lens,
        )
    else:
        raise ValueError((
            "Parse is only defined for 'dbcan', 'hmmscan3-domtab', "
            "'hmmer3-text'"
        ))


def decode_object(
    handle: BinaryIO,
    num_lines: Optional[float] = None,
) -> StringIO:
    """ Converts a binary file-like object into a String one.

    Files transferred over http arrive as binary objects.
    If I want to check the first few lines I need to decode it.
    """

    # If we don't specify a number of lines, just use infinity
    if num_lines is None:
        num_lines = float("inf")
    lines = [l.decode() for i, l
             in enumerate(handle)
             if i < num_lines]

    # Create an in memory file like object
    out = StringIO()
    # Write the lines to the file like object.
    out.writelines(lines)

    # Rewind the files to the start so that we can read them
    out.seek(0)
    handle.seek(0)
    return out


def split_hmm(string: str) -> str:
    return REGEX.sub("", str(string))


def parse_str_as_int(val: str, column: str) -> int:
    try:
        return int(val)
    except ValueError:
        msg = "Illegal value '{}' received in column {}. Expected an integer."
        raise LineParseError(msg.format(val, column))


def parse_str_as_float(val: str, column: str) -> float:
    try:
        return float(val)
    except ValueError:
        msg = "Illegal value '{}' received in column {}. Expected a float."
        raise LineParseError(msg.format(val, column))


class DBCAN(NamedTuple):
    """ Stores a single row of the DBCAN tab-delimited file. """

    hmm: str
    hmm_len: int
    seqid: str
    query_len: int
    evalue: float
    hmm_from: int
    hmm_to: int
    ali_from: int
    ali_to: int
    coverage: float

    @classmethod
    def from_string(cls, line: str) -> "DBCAN":
        """ Parse a single tabular output row from dbCAN. """

        columns = ["hmm", "hmm_len", "seqid", "query_len", "evalue",
                   "hmm_from", "hmm_to", "ali_from", "ali_to", "coverage"]

        sline = line.strip().split('\t')

        if len(columns) != len(sline):
            msg = "Wrong number of columns. Found {}, expected {}."
            raise LineParseError(msg.format(len(sline), len(columns)))

        output = dict(zip(columns, sline))
        hmm = split_hmm(output["hmm"])
        hmm_len = parse_str_as_int(output["hmm_len"], "hmm_len")
        seqid = output["seqid"]
        query_len = parse_str_as_int(output["query_len"], "query_len")
        evalue = parse_str_as_float(output["evalue"], "evalue")
        hmm_from = parse_str_as_int(output["hmm_from"], "hmm_from")
        hmm_to = parse_str_as_int(output["hmm_to"], "hmm_to")
        ali_from = parse_str_as_int(output["ali_from"], "ali_from")
        ali_to = parse_str_as_int(output["ali_to"], "ali_to")
        coverage = parse_str_as_float(output["coverage"], "coverage")

        return cls(hmm, hmm_len, seqid, query_len, evalue, hmm_from, hmm_to,
                   ali_from, ali_to, coverage)

    @classmethod
    def from_file(cls, handle: TextIO) -> Iterator["DBCAN"]:
        """ Parse a DBCAN file into generator of rows. """

        nmatches = 0
        for line_number, line in enumerate(handle, 1):
            line = line.strip()
            # Skip comment lines
            if line.startswith("#"):
                continue
            # Skip empty lines
            elif line == "":
                continue

            nmatches += 1
            try:
                parsed = cls.from_string(line)
            except LineParseError as e:
                predicted = predict_filetype(handle)
                if hasattr(handle, "name"):
                    name = handle.name
                else:
                    name = None

                if predicted != "dbcan":
                    msg = (
                        "Couldn't parse file as dbcan format.\n"
                        "Double check that the input file is in the right "
                        "format (i.e. dbcan vs hmmer_text vs hmmer_domtab).\n"
                        "Based on the file contents, we think this is in "
                        "'{}' format.\n"
                        "If you believe this is an error, please contact us."
                    ).format(predicted)
                    raise ParseError(name, None, msg)

                else:
                    raise ParseError(name, line_number, e.message)

            yield parsed

        if nmatches == 0:
            if hasattr(handle, "name"):
                name = handle.name
            else:
                name = None

            print(
                f"WARNING: input {name} has zero CAZymes detected.",
                file=sys.stderr
            )
            print(
                "WARNING: This will result in poor predictions.",
                file=sys.stderr
            )
            print(
                "WARNING: Please double check that you have "
                "specified the correct file format.",
                file=sys.stderr
            )
        return


class HMMER(NamedTuple):
    """ Stores a single HMMER match. """

    hmm: str
    hmm_len: int
    seqid: str
    evalue: float
    hmm_from: int
    hmm_to: int
    ali_from: int
    ali_to: int
    coverage: float

    @classmethod
    def from_file(  # noqa: C901
        cls,
        handle: TextIO,
        hmm_lens: Dict[str, int],
        format: str = "hmmer3-text",
    ) -> Iterator["HMMER"]:
        """ Processes HMMER output in a similar way to dbCAN hmmscan-parser.sh.

        There are two main HMMER output types, the plaintext output and the
        domain table type output (provided by `--domtblout`). The plaintext
        format is not particularly friendly, so we're using the Biopython
        parser.  This has the added benefit of handling both the domain table
        and plaintext format with minimal alteration.

        Keyword arguments:
        handle -- A python file handle of the HMMER output.
        format -- A string specifying an input format supported by the
            Biopython SearchIO API. Should be either 'hmmer3-text' or
            'hmmscan3-domtab'.

        Returns:
        generator -- A generator yielding named tuples corresponding to lines.
        """

        # This should never be raised at runtime, but avoids foot shooting
        # while writing code.
        assert format in ("hmmer3-text", "hmmscan3-domtab")

        # This is the threshold used by dbcan to determine if to matches are
        # the same.
        overlap_thres = 0.5

        qcounts = 0

        # Parse the file using Biopython.
        matches = SearchIO.parse(handle, format=format)

        # Loop through matches
        try:
            for query in matches:
                # For each HMM type that matches the sequence
                qcounts += 1

                query_matches = list()
                for hit in query.hits:
                    # For each partial alignment of the HMM to the sequence.
                    # Multiple HSPs can occur if there are genuinely two
                    # copies of the domain or if a large domain is split
                    # into two alignments.
                    for hsp in hit.hsps:
                        # Drop rows with alignment lengths of zero
                        if hsp.query_end - hsp.query_start <= 0:
                            continue

                        # Format and yield the line elements that we want.
                        row = cls._get_hsp(query, hit, hsp, hmm_lens)
                        query_matches.append(row)
                # Get the best match for overlapping matches.
                query_matches.sort(key=lambda x: (x.ali_from, x.ali_to))
                query_winners = cls._distinct_matches(
                    query_matches,
                    overlap_thres
                )
                # Yield the rows.
                for winner in query_winners:
                    # If this winner passes the coverage and evalue thresholds
                    # yield it.
                    if cls._filter_significant(winner):
                        yield winner

        except AssertionError:
            predicted = predict_filetype(handle)
            msg = (
                "Couldn't parse file as {} format.\n"
                "Double check that the input file is in the right format "
                "(i.e. dbcan vs hmmer_text vs hmmer_domtab).\n"
                "Based on the file contents, we think this is in "
                "'{}' format.\n"
                "If you believe this is an error, please contact us."
            ).format(format, predicted)

            if hasattr(handle, "name"):
                name = handle.name
            else:
                name = None

            raise ParseError(name, None, msg)

        except KeyError as e:
            msg = (
                "The file appears to be a searched against a different "
                "version of DBCAN than specified.\n"
                "The CAZyme group {} doesn't exist in this version.\n"
                "Please check the database version and contact us if you "
                "think this was a mistake."
            ).format(e.args)

            if hasattr(handle, "name"):
                name = handle.name
            else:
                name = None

            raise ParseError(name, None, msg)

        if qcounts == 0:
            if hasattr(handle, "name"):
                name = handle.name
            else:
                name = None

            predicted = predict_filetype(handle)
            if format != predicted:
                msg = (
                    "Couldn't parse file as {} format.\n"
                    "Double check that the input file is in the right format "
                    "(i.e. dbcan vs hmmer_text vs hmmer_domtab).\n"
                    "Based on the file contents, we think this is in "
                    "'{}' format.\n"
                    "If you believe this is an error, please contact us."
                ).format(format, predicted)
                raise ParseError(name, None, msg)
            else:
                print(
                    f"WARNING: input {name} has zero CAZymes detected.",
                    file=sys.stderr
                )
                print(
                    "WARNING: This will result in poor predictions.",
                    file=sys.stderr
                )
                print(
                    "WARNING: Please double check that you have "
                    "specified the correct file format.",
                    file=sys.stderr
                )
                print(
                    "WARNING: For HMMER output, try using the alternate "
                    "format e.g. 'hmmer_text' or 'hmmer_domtab'.",
                    file=sys.stderr
                )
        return

    @classmethod
    def _get_hsp(
        cls,
        query: QueryResult,
        hit: Hit,
        hsp: HSP,
        hmm_lens: Dict[str, int]
    ) -> "HMMER":
        """ Construct a record from a biopython SearchIO tree. """

        hmm = split_hmm(hit.id)
        hmm_len = hmm_lens[hmm]

        hmm_from = hsp.hit_start
        hmm_to = hsp.hit_end

        coverage = (hmm_to - hmm_from) / hmm_len
        return cls(
            hmm=hmm,
            hmm_len=hmm_len,
            seqid=query.id,
            evalue=hsp.evalue,
            hmm_from=hmm_from,
            hmm_to=hmm_to,
            ali_from=hsp.query_start,
            ali_to=hsp.query_end,
            coverage=coverage,
        )

    @staticmethod
    def _distinct_matches(
        matches: Sequence["HMMER"],
        overlap_thres: float
    ) -> List["HMMER"]:
        # Accumulate winners into a list.
        winners = list()

        # Keep track of the current best hits.
        winner = None
        winner_ali_length = 0

        # Loop through each match for each seqid, replacing "winner" as we go.
        for match in matches:
            # If this is the first match winner will be None, so we just set
            # it to the first value and continue.
            if winner is None:
                winner = match
                winner_ali_length = winner.ali_to - winner.ali_from
                # Continue moves to the next match in the list.
                continue

            # Get the distance between current match and winner
            dist = winner.ali_to - match.ali_from

            # Decide whether or not the matches are the same
            match_ali_length = match.ali_to - match.ali_from
            winner_nooverlap = (dist / winner_ali_length) > overlap_thres
            match_nooverlap = (dist / match_ali_length) > overlap_thres

            # If the distance is positive (meaning that row start is before
            # current winner end) and the distance is a large proportion of
            # the alignment length, consider them to be the same.
            if dist > 0 and (winner_nooverlap or match_nooverlap):
                # If the new rows evalue is less than the current winners
                # evalue the new row becomes the new winner!
                if match.evalue < winner.evalue:
                    winner = match
                    winner_ali_length = match_ali_length
            else:
                # We're onto a new alignment and the previous winner gets
                # to stay.
                winners.append(winner)
                winner = match
                winner_ali_length = match_ali_length

        # At the end we'll have a new winner to add.
        # But not if the list was empty.
        if winner is not None:
            winners.append(winner)
        return winners

    @staticmethod
    def _filter_significant(match: "HMMER") -> bool:
        """ Select significant alignments based on evalues and coverage.

        dbCAN uses a threshold of 1e-5 for alignments over 80 AA
        and 1e-3 for others. A coverage threshold of 0.3 is used.
        """

        if match.coverage < 0.3:
            return False

        is_long = (match.ali_to - match.ali_from) > 80

        if is_long:
            return match.evalue < 1e-5
        else:
            return match.evalue < 1e-3
